/*
 * lilJudd
 * Copyright (C) 2023 moonleay
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

package net.moonleay.lilJudd.features

import com.kotlindiscord.kord.extensions.utils.isNullOrBot
import dev.inmo.krontab.buildSchedule
import dev.inmo.krontab.doInfinity
import dev.kord.common.entity.Snowflake
import dev.kord.core.behavior.UserBehavior
import dev.kord.core.behavior.createRole
import dev.kord.core.behavior.requestMembers
import dev.kord.core.entity.channel.Channel
import dev.kord.core.entity.channel.MessageChannel
import dev.kord.gateway.PrivilegedIntent
import dev.kord.rest.builder.message.EmbedBuilder
import net.moonleay.lilJudd.Bot
import net.moonleay.lilJudd.data.database.entry.PlanningNotifierRolesData
import net.moonleay.lilJudd.data.database.entry.TimePlanningMessagesData
import net.moonleay.lilJudd.data.database.repository.PlanningNotifierRolesRepository
import net.moonleay.lilJudd.data.database.repository.TimePlanningMessagesRepository
import net.moonleay.lilJudd.extensions.FeatureManageExtension
import net.moonleay.lilJudd.features.component.FeatureEnum
import net.moonleay.lilJudd.features.component.IFeature
import net.moonleay.lilJudd.util.*
import java.time.ZonedDateTime

object AvailabilityManager : IFeature {

    // This runs during the cronjob.
    suspend fun runThread() {
        Logger.out("Starting to update roles...")

        // ChannelID, Data
        val messages = TimePlanningMessagesRepository.getWeek(TimeUtil.getWeekStamp().toEpochSecond())
            .associateBy { it.channelID }
        val targetedRoles = PlanningNotifierRolesRepository.getAll().associateBy { it.channelID }
        if (targetedRoles.isEmpty()) {
            Logger.out("No saved roles. Canceling.")
            return
        }

        for (id in messages.keys) {
            val snf = Snowflake(id) // snf = Snowflake
            val data = messages[id]!! // this is the data of the table
            if (Bot.bot.kordRef.getChannel(Snowflake(data.channelID)) == null) {
                // This channel does not exist anymore.
                Logger.out("Warning: Channel ${data.channelID} does not exist anymore. Skipping.")
                continue
            }
            val roleData = targetedRoles[data.channelID] // Get the role data
            if (roleData == null) {
                Logger.out("Role for channel ${data.channelID} does not exist")
                continue // this took way to long to find out that this was the issue
            }
            this.updateInChannel(snf, data, roleData)
        }
        Logger.out("Done! Until tomorrow! <3   ")
    }

    suspend fun updateInChannel(snf: Snowflake) {
        val stamp = TimeUtil.getWeekStamp().toEpochSecond()
        Logger.out("Weekstamp: $stamp")
        val messageData = TimePlanningMessagesRepository.getWeekInChannel(
            stamp,
            snf.value.toLong()
        )
        if (messageData == null) {
            Logger.out("Could not find data for channel ${snf.value}")
            return
        }
        val roleData = PlanningNotifierRolesRepository.getForChannel(snf.value.toLong())
        if (roleData == null) {
            Logger.out("Role for channel ${messageData.channelID} does not exist")
            return // this took way to long to find out that this was the issue
        }
        updateInChannel(snf, messageData, roleData)
    }

    @OptIn(PrivilegedIntent::class)
    suspend fun updateInChannel(snf: Snowflake, tpmd: TimePlanningMessagesData, pnrd: PlanningNotifierRolesData) {
        if (Bot.bot.kordRef.getChannel(snf) == null)
            return // This channel does not exist anymore.
        val c = Bot.bot.kordRef.getChannelOf<MessageChannel>(snf)!! // Get the channel as MessageChannel
        val weekday = ZonedDateTime.now().dayOfWeek // The current week day
        val weekStamp = TimeUtil.getWeekStamp().toEpochSecond() // The current week time stamp
        Logger.out("It is week ${weekStamp} and day ${weekday}/${TimeUtil.getDayOfMonthInt(weekday)} of the week.")
        val g = Bot.bot.kordRef.getGuild(Snowflake(tpmd.serverID))
        // Get all members with the role
        val mce = g.requestMembers {
            this.requestAllMembers()
        }
        mce.collect { memberchunkevent ->
            memberchunkevent.members.forEach {
                if (!it.isNullOrBot()) { // Check if the member is a bot
                    Logger.out("Checking member ${it.id.value} (${it.username})")
                    if (it.roleIds.contains(Snowflake(pnrd.hasTimeRoleID))) {
                        it.removeRole(Snowflake(pnrd.hasTimeRoleID))
                        Logger.out("Removed role from ${it.username}") // Removed the role
                    }
                }
                // I cant use continue here, because it does not work with .forEach
            }
        }

        Logger.out("Got through all members")
        // This stores the ids of the messages.
        // The format is weekdaNR:ID
        // The last entry (nr 8) is empty, so we can ignore it
        val messageIdSplit = tpmd.messageIDs.split(";").subList(0, 7)
        for (mid in messageIdSplit) {
            Logger.out("Checking id $mid")
            if (!mid.startsWith((TimeUtil.getDayOfMonthInt(weekday) - 1).toString(), true))
                continue// This is not the right message, check the next one
            val idFiltered = mid.split(":")[1] // This is the target message id
            val message = c.getMessageOrNull(Snowflake(idFiltered)) // Get the message from the channel
            if (message == null) {
                Logger.out("Could not find message.")
                return // This message does not exist anymore. Nothing we can do about that.
            }
            if (message.data.embeds.isEmpty()) {
                Logger.out("There are no embeds.")
                return // There are no embeds or there are not enough embeds
            }
            val targets = EmbedUtil.getAllUsersInTheFirstXTables(2, message.embeds[0])
            for (tid in targets) {
                Logger.out("Checking id $tid")
                if (Bot.bot.kordRef.getGuildOrNull(Snowflake(tpmd.serverID))!!
                        .getMemberOrNull(Snowflake(tid)) == null
                )
                    continue// This member does not exist anymore.
                val member = Bot.bot.kordRef.getGuild(Snowflake(tpmd.serverID))
                    .getMember(Snowflake(tid)) // Get the member
                if (member.roleIds.contains(Snowflake(pnrd.hasTimeRoleID)))
                    continue // This member already has the role
                member.addRole(Snowflake(pnrd.hasTimeRoleID)) // Add the role
                Logger.out("Added role to ${member.username}")
            }
            Logger.out("Done with message. Moving on...")
            // We found the right message. We don't need to check the others.
            break
        }
    }

    override val feat: FeatureEnum
        get() = FeatureEnum.PLANNINGROLES

    // Register the cronjob to run at 1AM UTC every day
    override suspend fun registerThread() {
        Logger.out("Adding availability scheduler...")
        val scheduler = buildSchedule("0 0 2 * * *") // 0 0 4 * * * 0o 1w // 0o is UTC
        scheduler.doInfinity {
            this.runThread()
        }
    }

    override suspend fun enable(
        u: UserBehavior,
        gID: Long,
        cID: Long,
        ch: Channel,
        args: FeatureManageExtension.FeatureManagerArgs
    ): EmbedBuilder {
        var alreadyExists = PlanningNotifierRolesRepository.existsInChannel(cID)
        // Check if the channel and guild already exist in the db
        if (!alreadyExists) {
            // Create the roles in Discord
            val hasTimeRole = Bot.bot.kordRef.getGuild(Snowflake(gID)).createRole {
                this.name = "available [${ch.data.name.value}]"
                this.mentionable = true
            }
            val htr = hasTimeRole.id.value.toLong()

            val wantsNotifsRole = Bot.bot.kordRef.getGuild(Snowflake(gID)).createRole {
                this.name = "notifications [${ch.data.name.value}]"
                this.mentionable = true
            }
            val wnr = wantsNotifsRole.id.value.toLong()

            // Save the role ids to db
            PlanningNotifierRolesRepository.write(
                PlanningNotifierRolesData(
                    id = -1,
                    serverID = gID,
                    channelID = cID,
                    hasTimeRoleID = htr,
                    wantsToBeNotifiedID = wnr
                )
            )

            return MessageUtil.getEmbed(
                EmbedColor.SUCCESS,
                "200: Success",
                "The feature was enabled in channel ${args.channel.data.name.value} with roles ${hasTimeRole.mention} & ${wantsNotifsRole.mention}.",
                u.asUser().username + "#" + u.asUser().discriminator
            )
        }

        // They exist, do not add them
        return MessageUtil.getEmbed(
            EmbedColor.ERROR,
            "403: Forbidden",
            "The feature is already enabled in this channel.",
            u.asUser().username + "#" + u.asUser().discriminator
        )
    }

    override suspend fun disable(
        u: UserBehavior,
        gID: Long,
        cID: Long,
        ch: Channel,
        args: FeatureManageExtension.FeatureManagerArgs
    ): EmbedBuilder {
        // Check if entry exists in db
        if (PlanningNotifierRolesRepository.existsInChannelFromSever(cID, gID)) {
            val entry = PlanningNotifierRolesRepository.getForChannelInServer(cID, gID)!!
            // delete all entries for this guild and channel combo
            Bot.bot.kordRef.getGuild(Snowflake(gID))
                .getRoleOrNull(Snowflake(entry.hasTimeRoleID))?.delete()
            Bot.bot.kordRef.getGuild(Snowflake(gID))
                .getRoleOrNull(Snowflake(entry.wantsToBeNotifiedID))?.delete()

            // delete all found entries
            PlanningNotifierRolesRepository.delete(entry.id)
            return MessageUtil.getEmbed(
                EmbedColor.SUCCESS,
                "200: Success",
                "The feature was disabled.",
                u.asUser().username + "#" + u.asUser().discriminator
            )
        }
        // not in db, do nothing
        return MessageUtil.getEmbed(
            EmbedColor.ERROR,
            "403: Forbidden",
            "The feature is already disabled in this channel.",
            u.asUser().username + "#" + u.asUser().discriminator
        )
    }
}
